#include "BF_MSNR.h"

#include <stdio.h>
#include <iostream>

namespace BF
{
    CMSNR::CMSNR(int _numFreq, int _lowerfreq, int _upperfreq, int _samplingRate, double _soundSpeed, std::string _decomp_alg, double _alpha) :
	BF(_numFreq, _lowerfreq, _upperfreq, _samplingRate, _soundSpeed, UtID),
	decomp_alg(_decomp_alg), alpha(_alpha)
    {
		this->iniRnnFlag = 0; /* add */
    }

    CMSNR::~CMSNR()
    {
    }

    void
    CMSNR::updateSources(const std::vector<SourceLocation> &sourcePositions)
    {
/*== source from adaptBF.m

  % This algorithm requires VAD data (Adata.vad(NSrc), = 0(noise) or 1(signal))
  % This algorithm uses 2 intermediate variables Rss, Rnn, mean 
  % correlation matrixes of signal(incl. noise) and noise, respectively.
  % The freedom of scaling is fixed by |g| = 1. (g is BF coef. vector)
  % This algorithm requires to set initial G for getting size information.
  % To set zero matrix for G is OK.
      if isempty(Adata.Rss) % Initialization process
        Adata.Rss = zeros(NMic, NMic, NSrc, NFreq); % Signal correlation matrix
        Adata.Rnn = zeros(NMic, NMic, NSrc, NFreq); % Noise correlation matrix
        for nf = 1:NFreq
          for ns = 1:NSrc % for each source
            Adata.Rss(:, :, ns, nf) = eye(NMic);
            Adata.Rnn(:, :, ns, nf) = eye(NMic);
            G(ns, :, nf) = ones(NMic, 1)./sqrt(NMic); % Initial BF coef.
          end
        end
        return
      else % update process
        for ns = 1:NSrc % for each source
          vad = Adata.vad(ns);
          for nf = 1:NFreq
              x = Adata.x(:, nf);
              Rxx = x*x';
              alp = Adata.alp;
              Rss = Adata.Rss(:, :, ns, nf);
              Rnn = Adata.Rnn(:, :, ns, nf);
              if vad == 1 % Signal part
                  Rss = alp * Rss + (1-alp) * Rxx;
                  Adata.Rss(:, :, ns, nf) = Rss;
              else % Noise part
                  Rnn = alp * Rnn + (1-alp) * Rxx;
                  Adata.Rnn(:, :, ns, nf) = Rnn;
              end
              [V, D] = eig(Rss, Rnn);
              [msnr, mi] = max(diag(D));
              g = V(:, mi);
              g = g ./ sqrt(real(g'*g));
              G(ns, :, nf) = g'; % Store total BF coef.
          end
        end
        return
      end % For if initialization
      
*/
	const int numSrc = sourcePositions.size();
	const int numMic = this->numMicChannels;

	/*== 直近フレームでの音源と、このフレームでの音源との対応をとる ==*/
	std::vector<int> lastSourceIndices; // i 番めの要素は、このフレームでの i 番目の音源（配列の i 番目の要素という意味）が、直近フレームでの何番目の音源（同）に相当するかを表す。
	for(int i = 0; i < numSrc; i++){
	    const SourceLocation &sourcePosition = sourcePositions[i];

	    /* 直近フレームの音源から、同じ ID を持つものを探す */
	    int j;
		for(j = 0; j < this->sourcePositions.size(); j++){
			if(sourcePosition.sourceID == this->sourcePositions[j].sourceID){
				break;
			}
	    }
	    
	    if(j < this->sourcePositions.size()){
			/* 見つかった */
			lastSourceIndices.push_back(j);
	    }else{
			/* 見つからなかった */
			lastSourceIndices.push_back(-1);
	    }
	}

	/*== 伝達関数・分離行列を再計算する ==*/
	std::vector<Matrix> newGs;
	std::vector<Matrix> newRss, newG;
	for (int i = 0; i < this->numFreq; i++) {
	    // 対象周波数のみ計算
	    if (i < lowerfreq || i > upperfreq){
	    	newGs.push_back(Matrix::O(numSrc, numMic));
	        for(int j = 0; j < numSrc; j++){
		    	newRss.push_back(Matrix::I(numMic));
			}
	    	continue;
	    }

	    Matrix newG(numSrc, numMic);
	    for (int j = 0; j < numSrc; j++) {
			/* 入力の j 番目の音源と同じIDの音源が、直近フレームに見当たらない？ */
			const int lastSourceIndex = lastSourceIndices[j];
			if (lastSourceIndex < 0) {
				newRss.push_back(Matrix::I(numMic));
//if (i == 0 && j == 0) {std::cout << "Rss initialized." << std::endl;}

				for(int k = 0; k < numMic; k++){
					newG.setElementAt(j, k, 1.0 / sqrt(numMic));
				}
			} else {
				newRss.push_back(this->Rss[i * this->sourcePositions.size() + lastSourceIndex]);

			    newG.setRow(j, this->separationMatrices[i].getRow(lastSourceIndex));
			}
		}
		newGs.push_back(newG);
	}
	/*== 結果の格納 ==*/
	this->sourcePositions = sourcePositions;
	newRss.swap(this->Rss);
	newGs.swap(this->separationMatrices);
    }


    void
    CMSNR::adaptSeparationMatrices()
    {
	/*const int numSrc = this->sourcePositions.size();*/
	const int numSrc = this->numSrcPos; /* update */

	//const int numMic = this->transformFunctions[0].getRowCount();
	const int numMic = this->numMicChannels;
	

	// 最初の一回だけ、ここで、Rnnを初期化する /* add */
	if (this->iniRnnFlag == 0) {
//std::cout << "Rnn initialized." << std::endl;
		std::vector<Matrix> newRnn;
		for (int i = 0; i < this->numFreq; i++) {
			newRnn.push_back(Matrix::I(numMic));
		}
		newRnn.swap(this->Rnn);
		this->iniRnnFlag = 1;
	}


	std::vector<Matrix> newRss, newRnn, newGs;
	double alpha = this->alpha;
	for(int i = 0; i < this->numFreq; i++){
		// 対象周波数のみを更新
		if (i < lowerfreq || i > upperfreq) {
			if (numSrc > 0) {
				for (int j = 0; j < numSrc; j++) {
					newRss.push_back(Matrix::I(numMic));
				}
				newGs.push_back(Matrix::O(numSrc, numMic));
			}
			newRnn.push_back(Matrix::I(numMic));
			continue;
		}

	    /*for(int j = 0; j < numSrc; j++){
			bool isSignal = this->sourcePositions[j].isSignal;
			if(isSignal){
				...
			}else{
				...
			}
		}*//* update */

		/*Matrix x     = this->inputSpectrum.getColumn(i);*/
		/*Matrix x     = this->inputSpectrum.getColumn(i) * (1.0 / 32768.0);*/
		Matrix x     = this->inputSpectrum.getColumn(i) / 32768.0;//8388608.0
		Matrix rxx   = x * x.transposed();
		
		if (numSrc > 0) {
			Matrix rnn;
			rnn = this->Rnn[i];

		    Matrix newG(numSrc, numMic);
			for (int j = 0; j < numSrc; j++) {
				Matrix rss;
				rss = alpha * this->Rss[i * numSrc + j] + (1.0 - alpha) * rxx;
				newRss.push_back(rss);

				Matrix evec, eval, right_evec;
				/* Rnn が I でない場合、octave の結果とこの一般固有値の計算結果は一致しないようである。
				 * ただし、octave の結果は AV-BVD が 0 ではないので、octave 実装に誤差があると予想する。
				 */
				if(decomp_alg == "GEVD")
				  rss.generalizedEigenValues(rnn, eval, evec);
				else if(decomp_alg == "GSVD")
				  rss.generalizedSingularValues(rnn, evec, eval, right_evec);

				int m = eval.getMaximumValueIndexInColumn(0);
				Matrix g = evec.getColumn(m);
				/*g = 1.0 / g.norm().real() * g;*/
				//g = g * (1.0 / sqrt(g.norm().real())); /* update */
				g = g / g.norm(); /* update */
//g.dump(stdout, "g");
				newG.setRow(j, g.transposed());
//if (i == 0) std::cout << "Rss[" << j << "] adapted." << std::endl;
			}
			newGs.push_back(newG);
		}
		else {
			Matrix rnn;
			rnn = alpha * this->Rnn[i] + (1.0 - alpha) * rxx;
			newRnn.push_back(rnn);
//if (i == 0) std::cout << "Rnn adapted." << std::endl;
		}
	}

	/*== 結果の格納 ==*/
	if (numSrc > 0) {
		newRss.swap(this->Rss);
		newGs.swap(this->separationMatrices);
	}
	else {
		newRnn.swap(this->Rnn);
	}
    }
}
